@@ -1,4 +1,5 @@
 use std::collections::HashSet;
+use std::env;
 use std::env::consts::ARCH;
 use std::ffi::OsStr;
 use std::ffi::OsString;
@@ -54,6 +55,7 @@ fn extract_version(output: &str) -> Result<&str> {
 /// Directory and enclosed contents will be removed when return object is dropped.
 #[cfg(feature = "default")]
 fn extract_libbpf_headers_to_disk(target_dir: &Path) -> Result<Option<PathBuf>> {
+    use libbpf_rs::libbpf_sys;
     use std::fs::OpenOptions;
     use std::io::Write;
 
@@ -183,12 +185,17 @@ fn compile_one(
     {
         // We may end up being invoked by a build script, in which case
         // `CARGO_CFG_TARGET_ARCH` would represent the target architecture.
-        let arch = option_env!("CARGO_CFG_TARGET_ARCH").unwrap_or(ARCH);
+        let arch = env::var("CARGO_CFG_TARGET_ARCH");
+        let arch = arch.as_deref().unwrap_or(ARCH);
         let arch = match arch {
             "x86_64" => "x86",
             "aarch64" => "arm64",
             "powerpc64" => "powerpc",
             "s390x" => "s390",
+            "riscv64" => "riscv",
+            "loongarch64" => "loongarch",
+            "sparc64" => "sparc",
+            "mips64" => "mips",
             x => x,
         };
         cmd.arg(format!("-D__TARGET_ARCH_{arch}"));
