@@ -2,6 +2,9 @@ use std::borrow::Cow;
 use std::cell::RefCell;
 use std::collections::HashMap;
 use std::collections::HashSet;
+use std::fmt::Debug;
+use std::fmt::Formatter;
+use std::fmt::Result as FmtResult;
 use std::fmt::Write;
 use std::mem::size_of;
 use std::num::NonZeroUsize;
@@ -10,6 +13,7 @@ use std::ops::Deref;
 use anyhow::anyhow;
 use anyhow::bail;
 use anyhow::ensure;
+use anyhow::Context as _;
 use anyhow::Result;
 use libbpf_rs::btf::types;
 use libbpf_rs::btf::types::Linkage;
@@ -250,8 +254,27 @@ fn escape_reserved_keyword(identifier: Cow<'_, str>) -> Cow<'_, str> {
         "please keep reserved keywords sorted",
     );
 
+    // Some keywords just cannot be used, even in raw form.
+    // See https://internals.rust-lang.org/t/raw-identifiers-dont-work-for-all-identifiers/9094
+    let disallowed_raw = ["Self", "crate", "self", "super"];
+    debug_assert_eq!(
+        disallowed_raw.as_slice(),
+        {
+            let mut vec = disallowed_raw.to_vec();
+            vec.sort();
+            vec
+        },
+        "please keep reserved keywords sorted",
+    );
+
     if reserved.binary_search(&identifier.as_ref()).is_ok() {
-        Cow::Owned(format!("r#{identifier}"))
+        if disallowed_raw.binary_search(&identifier.as_ref()).is_ok() {
+            // Just remove the first 'a' or 'e' character. Yes, that could
+            // conceivably be the cause of a collision in itself ¯\_(ツ)_/¯
+            Cow::Owned(identifier.replacen(['a', 'e'], "", 1))
+        } else {
+            Cow::Owned(format!("r#{identifier}"))
+        }
     } else {
         identifier
     }
@@ -278,11 +301,168 @@ impl AnonTypes {
     }
 }
 
+
+#[derive(Debug)]
+pub(crate) struct GenStructOps<'btf> {
+    btf: &'btf GenBtf<'btf>,
+    deps: Vec<BtfType<'btf>>,
+    vars: Vec<types::Var<'btf>>,
+}
+
+impl<'btf> GenStructOps<'btf> {
+    pub fn new(btf: &'btf GenBtf<'btf>) -> Result<Self> {
+        let mut deps = Vec::new();
+        let mut vars = Vec::new();
+
+        // Take all the struct_ops datasec entries and collect their variables
+        // (and dependent types).
+        for ty in btf.type_by_kind::<types::DataSec<'_>>() {
+            let name = match ty.name() {
+                Some(s) => s.to_str().context("datasec has invalid name")?,
+                None => "",
+            };
+
+            if !matches!(
+                canonicalize_internal_map_name(name),
+                Some(InternalMapType::StructOps)
+            ) {
+                continue;
+            }
+
+            for var in ty.iter() {
+                let var = btf
+                    .type_by_id::<types::Var<'_>>(var.ty)
+                    .ok_or_else(|| anyhow!("datasec type does not point to a variable"))?;
+
+                if var.linkage() == types::Linkage::Static {
+                    // do not output Static Var
+                    continue;
+                }
+
+                let () = vars.push(var);
+
+                if let Some(next_ty) = next_type(*var)? {
+                    let () = deps.push(next_ty);
+                }
+            }
+        }
+
+        let slf = Self { btf, deps, vars };
+        Ok(slf)
+    }
+
+    pub fn gen_struct_ops_def(&self, def: &mut String) -> Result<()> {
+        // Emit a single struct_ops type definition containing all
+        // variables discovered earlier.
+        write!(
+            def,
+            r#"
+#[derive(Debug, Clone)]
+#[repr(C)]
+pub struct StructOps {{
+"#
+        )?;
+
+        for var in self.vars.iter() {
+            writeln!(
+                def,
+                r#"    pub {var_name}: *mut types::{var_type},"#,
+                var_name = var.name().unwrap().to_string_lossy(),
+                var_type = self.btf.type_declaration(**var)?
+            )?;
+        }
+
+        writeln!(def, "}}")?;
+
+        write!(
+            def,
+            r#"
+impl StructOps {{
+"#
+        )?;
+
+        for var in self.vars.iter() {
+            write!(
+                def,
+                r#"
+    pub fn {var_name}(&self) -> &types::{var_type} {{
+        // SAFETY: The library ensures that the member is pointing to
+        //         valid data.
+        unsafe {{ self.{var_name}.as_ref() }}.unwrap()
+    }}
+
+    pub fn {var_name}_mut(&mut self) -> &mut types::{var_type} {{
+        // SAFETY: The library ensures that the member is pointing to
+        //         valid data.
+        unsafe {{ self.{var_name}.as_mut() }}.unwrap()
+    }}
+"#,
+                var_name = var.name().unwrap().to_string_lossy(),
+                var_type = self.btf.type_declaration(**var)?
+            )?;
+        }
+
+        writeln!(def, "}}")?;
+        Ok(())
+    }
+
+    pub fn gen_dependent_types(
+        mut self,
+        processed: &mut HashSet<TypeId>,
+        def: &mut String,
+    ) -> Result<()> {
+        let vars = self
+            .vars
+            .iter()
+            .map(|ty| ty.next_type().unwrap().type_id())
+            .collect::<HashSet<_>>();
+
+        while !self.deps.is_empty() {
+            let ty = self.deps.remove(0);
+            if !processed.insert(ty.type_id()) {
+                continue;
+            }
+
+            btf_type_match!(match ty {
+                BtfKind::Composite(t) => {
+                    if vars.contains(&ty.type_id()) {
+                        let opts = TypeDeclOpts {
+                            func_type: "libbpf_rs::libbpf_sys::bpf_program",
+                        };
+                        self.btf.type_definition_for_composites_with_opts(
+                            def,
+                            &mut self.deps,
+                            t,
+                            &opts,
+                        )?
+                    } else {
+                        self.btf
+                            .type_definition_for_composites(def, &mut self.deps, t)?
+                    }
+                }
+                BtfKind::Enum(t) => self.btf.type_definition_for_enums(def, t)?,
+                _ => bail!("Invalid type: {:?}", ty.kind()),
+            });
+        }
+
+        Ok(())
+    }
+}
+
+
 pub struct GenBtf<'s> {
     btf: Btf<'s>,
     anon_types: AnonTypes,
 }
 
+impl Debug for GenBtf<'_> {
+    fn fmt(&self, f: &mut Formatter<'_>) -> FmtResult {
+        f.debug_struct("GenBtf<'_>")
+            .field("btf", &self.btf)
+            .finish()
+    }
+}
+
 impl<'s> From<Btf<'s>> for GenBtf<'s> {
     fn from(btf: Btf<'s>) -> GenBtf<'s> {
         Self {
@@ -350,10 +530,9 @@ impl<'s> GenBtf<'s> {
             )
         };
 
-        ensure!(
-            !is_terminal(ty),
-            "Tried to print type definition for terminal type"
-        );
+        if is_terminal(ty) {
+            return Ok(String::new());
+        }
 
         // Process dependent types until there are none left.
         //
@@ -380,130 +559,6 @@ impl<'s> GenBtf<'s> {
         Ok(def)
     }
 
-    pub fn struct_ops_type_definition(&self, processed: &mut HashSet<TypeId>) -> Result<String> {
-        let mut def = String::new();
-        let mut dependent_types = vec![];
-        let mut vars = vec![];
-
-        // Take all the struct_ops datasec entries and collect their variables
-        // (and dependent types).
-        for ty in self.type_by_kind::<types::DataSec<'_>>() {
-            let name = match ty.name() {
-                Some(s) => s.to_str()?,
-                None => "",
-            };
-
-            if !matches!(
-                canonicalize_internal_map_name(name),
-                Some(InternalMapType::StructOps)
-            ) {
-                continue;
-            }
-
-            for var in ty.iter() {
-                let var = self
-                    .type_by_id::<types::Var<'_>>(var.ty)
-                    .ok_or_else(|| anyhow!("datasec type does not point to a variable"))?;
-
-                if var.linkage() == types::Linkage::Static {
-                    // do not output Static Var
-                    continue;
-                }
-
-                let () = vars.push(*var);
-
-                if let Some(next_ty) = next_type(*var)? {
-                    let () = dependent_types.push(next_ty);
-                }
-            }
-        }
-
-        // Emit a single struct_ops definition containing all variables
-        // discovered earlier.
-        write!(
-            def,
-            r#"
-#[derive(Debug, Clone)]
-#[repr(C)]
-pub struct struct_ops {{
-"#
-        )?;
-
-        for var in vars.iter() {
-            writeln!(
-                def,
-                r#"    pub {var_name}: *mut {var_type},"#,
-                var_name = var.name().unwrap().to_string_lossy(),
-                var_type = self.type_declaration(*var)?
-            )?;
-        }
-
-        writeln!(def, "}}")?;
-
-        write!(
-            def,
-            r#"
-impl struct_ops {{
-"#
-        )?;
-
-        for var in vars.iter() {
-            write!(
-                def,
-                r#"
-    pub fn {var_name}(&self) -> &{var_type} {{
-        // SAFETY: The library ensures that the member is pointing to
-        //         valid data.
-        unsafe {{ self.{var_name}.as_ref() }}.unwrap()
-    }}
-
-    pub fn {var_name}_mut(&mut self) -> &mut {var_type} {{
-        // SAFETY: The library ensures that the member is pointing to
-        //         valid data.
-        unsafe {{ self.{var_name}.as_mut() }}.unwrap()
-    }}
-"#,
-                var_name = var.name().unwrap().to_string_lossy(),
-                var_type = self.type_declaration(*var)?
-            )?;
-        }
-
-        writeln!(def, "}}")?;
-
-        let vars = vars
-            .into_iter()
-            .map(|ty| ty.next_type().unwrap().type_id())
-            .collect::<HashSet<_>>();
-
-        while !dependent_types.is_empty() {
-            let ty = dependent_types.remove(0);
-            if !processed.insert(ty.type_id()) {
-                continue;
-            }
-
-            btf_type_match!(match ty {
-                BtfKind::Composite(t) => {
-                    if vars.contains(&ty.type_id()) {
-                        let opts = TypeDeclOpts {
-                            func_type: "libbpf_rs::libbpf_sys::bpf_program",
-                        };
-                        self.type_definition_for_composites_with_opts(
-                            &mut def,
-                            &mut dependent_types,
-                            t,
-                            &opts,
-                        )?
-                    } else {
-                        self.type_definition_for_composites(&mut def, &mut dependent_types, t)?
-                    }
-                }
-                BtfKind::Enum(t) => self.type_definition_for_enums(&mut def, t)?,
-                _ => bail!("Invalid type: {:?}", ty.kind()),
-            });
-        }
-
-        Ok(def)
-    }
     fn type_definition_for_composites<'a>(
         &'a self,
         def: &mut String,
@@ -523,6 +578,11 @@ impl struct_ops {{
         t: types::Composite<'_>,
         opts: &TypeDeclOpts,
     ) -> Result<()> {
+        if t.is_empty_union() {
+            // Ignore empty union; Rust does not allow unions with no fields.
+            return Ok(());
+        }
+
         let packed = is_struct_packed(&t, &self.btf)?;
 
         // fields in the aggregate
@@ -547,6 +607,14 @@ impl struct_ops {{
                 .type_by_id::<BtfType<'_>>(member.ty)
                 .unwrap()
                 .skip_mods_and_typedefs();
+
+            if let Ok(composite) = TryInto::<types::Composite<'_>>::try_into(field_ty) {
+                if composite.is_empty_union() {
+                    // Skip empty union field; we do not generate a type for them.
+                    continue;
+                }
+            }
+
             if let Some(next_ty_id) = next_type(field_ty)? {
                 dependent_types.push(next_ty_id);
             }
@@ -575,12 +643,20 @@ impl struct_ops {{
                     impl_default.push(format!(
                         r#"            __pad_{offset}: [u8::default(); {padding}]"#,
                     ));
+
+                    if padding > 32 {
+                        gen_impl_default = true;
+                    }
                 }
 
                 if let Some(ft) = self.type_by_id::<types::Array<'_>>(field_ty.type_id()) {
                     if ft.capacity() > 32 {
                         gen_impl_default = true
                     }
+
+                    if self.type_by_id::<types::Ptr<'_>>(ft.ty()).is_some() {
+                        gen_impl_default = true
+                    }
                 }
 
                 // Rust does not implement `Default` for pointers, no matter if
@@ -634,6 +710,10 @@ impl struct_ops {{
                 impl_default.push(format!(
                     r#"            __pad_{offset}: [u8::default(); {padding}]"#,
                 ));
+
+                if padding > 32 {
+                    gen_impl_default = true;
+                }
             }
         }
 
@@ -669,11 +749,7 @@ impl struct_ops {{
                 self.anon_types.type_name_or_anon(&t),
             )?;
             writeln!(def, r#"    fn default() -> Self {{"#)?;
-            writeln!(
-                def,
-                r#"        {} {{"#,
-                self.anon_types.type_name_or_anon(&t)
-            )?;
+            writeln!(def, r#"        Self {{"#,)?;
             for impl_def in impl_default {
                 writeln!(def, r#"{impl_def},"#)?;
             }
@@ -702,11 +778,7 @@ impl struct_ops {{
                 self.anon_types.type_name_or_anon(&t),
             )?;
             writeln!(def, r#"    fn default() -> Self {{"#)?;
-            writeln!(
-                def,
-                r#"        {} {{"#,
-                self.anon_types.type_name_or_anon(&t)
-            )?;
+            writeln!(def, r#"        Self {{"#,)?;
             writeln!(def, r#"{},"#, impl_default[0])?;
             writeln!(def, r#"        }}"#)?;
             writeln!(def, r#"    }}"#)?;
@@ -766,14 +838,23 @@ impl struct_ops {{
         dependent_types: &mut Vec<BtfType<'a>>,
         t: types::DataSec<'_>,
     ) -> Result<()> {
-        let mut sec_name = match t.name().map(|s| s.to_string_lossy()) {
+        let sec_name = match t.name().map(|s| s.to_string_lossy().into_owned()) {
             None => bail!("Datasec name is empty"),
-            Some(s) if !s.starts_with('.') => bail!("Datasec name is invalid: {s}"),
-            Some(s) => s.into_owned(),
+            Some(mut s) if s.starts_with('.') => {
+                s.remove(0);
+                s
+            }
+            Some(s) => s,
         };
-        sec_name.remove(0);
         let sec_name = sec_name.replace('.', "_");
 
+        // Don't generate anything for ksyms. The BTF is patched up by libbpf at
+        // load time and the result can contain multiple variables with the same
+        // name, which is could result in invalid generated code.
+        if sec_name == "ksyms" {
+            return Ok(())
+        }
+
         writeln!(def, r#"#[derive(Debug, Copy, Clone)]"#)?;
         writeln!(def, r#"#[repr(C)]"#)?;
         writeln!(def, r#"pub struct {sec_name} {{"#)?;
