@@ -21,6 +21,7 @@ use tempfile::TempDir;
 
 use crate::build::build;
 use crate::gen::btf::GenBtf;
+use crate::gen::btf::GenStructOps;
 use crate::make::make;
 use crate::SkeletonBuilder;
 
@@ -376,84 +377,7 @@ fn test_make_workspace() {
         .exists());
 }
 
-#[test]
-fn test_skeleton_empty_source() {
-    let (_dir, proj_dir, cargo_toml) = setup_temp_project();
-
-    // Add prog dir
-    create_dir(proj_dir.join("src/bpf")).expect("failed to create prog dir");
-
-    // Add a prog
-    let _prog_file =
-        File::create(proj_dir.join("src/bpf/prog.bpf.c")).expect("failed to create prog file");
-
-    make(
-        true,
-        Some(&cargo_toml),
-        None,
-        Vec::new(),
-        true,
-        true,
-        Vec::new(),
-        None,
-    )
-    .unwrap();
-
-    let mut cargo = OpenOptions::new()
-        .append(true)
-        .open(&cargo_toml)
-        .expect("failed to open Cargo.toml");
-
-    // Make test project use our development libbpf-rs version
-    writeln!(
-        cargo,
-        r#"
-        libbpf-rs = {{ path = "{}" }}
-        "#,
-        get_libbpf_rs_path().as_path().display()
-    )
-    .expect("failed to write to Cargo.toml");
-
-    let mut source = OpenOptions::new()
-        .write(true)
-        .truncate(true)
-        .open(proj_dir.join("src/main.rs"))
-        .expect("failed to open main.rs");
-
-    write!(
-        source,
-        r#"
-        #![warn(elided_lifetimes_in_paths)]
-        mod bpf;
-        use bpf::*;
-        use libbpf_rs::skel::SkelBuilder;
-        use libbpf_rs::skel::OpenSkel;
-
-        fn main() {{
-            let builder = ProgSkelBuilder::default();
-            let _skel = builder
-                .open()
-                .expect("failed to open skel")
-                .load()
-                .expect("failed to load skel");
-        }}
-        "#,
-    )
-    .expect("failed to write to main.rs");
-
-    let status = Command::new("cargo")
-        .arg("build")
-        .arg("--quiet")
-        .arg("--manifest-path")
-        .arg(cargo_toml.into_os_string())
-        .env("RUSTFLAGS", "-Dwarnings")
-        .status()
-        .expect("failed to spawn cargo-build");
-    assert!(status.success());
-}
-
-#[test]
-fn test_skeleton_basic() {
+fn build_rust_project_from_bpf_c(bpf_c: &str, rust: &str) {
     let (_dir, proj_dir, cargo_toml) = setup_temp_project();
 
     // Add prog dir
@@ -466,28 +390,8 @@ fn test_skeleton_basic() {
         .truncate(true)
         .open(proj_dir.join("src/bpf/prog.bpf.c"))
         .expect("failed to open prog.bpf.c");
-
-    write!(
-        prog,
-        r#"
-        #include "vmlinux.h"
-        #include <bpf/bpf_helpers.h>
-
-        struct {{
-                __uint(type, BPF_MAP_TYPE_HASH);
-                __uint(max_entries, 1024);
-                __type(key, u32);
-                __type(value, u64);
-        }} mymap SEC(".maps");
-
-        SEC("kprobe/foo")
-        int this_is_my_prog(u64 *ctx)
-        {{
-                return 0;
-        }}
-        "#,
-    )
-    .expect("failed to write prog.bpf.c");
+    prog.write_all(bpf_c.as_bytes())
+        .expect("failed to write prog.bpf.c");
 
     // Lay down the necessary header files
     add_vmlinux_header(&proj_dir);
@@ -525,124 +429,135 @@ fn test_skeleton_basic() {
         .open(proj_dir.join("src/main.rs"))
         .expect("failed to open main.rs");
 
-    write!(
-        source,
-        r#"
+    source
+        .write_all(rust.as_bytes())
+        .expect("failed to write to main.rs");
+
+    let status = Command::new("cargo")
+        .arg("build")
+        .arg("--quiet")
+        .arg("--manifest-path")
+        .arg(cargo_toml.into_os_string())
+        .env("RUSTFLAGS", "-Dwarnings")
+        .status()
+        .expect("failed to spawn cargo-build");
+    assert!(status.success());
+}
+
+#[test]
+fn test_skeleton_empty_source() {
+    let bpf_c = String::new();
+    let rust = r#"
         #![warn(elided_lifetimes_in_paths)]
         mod bpf;
+        use std::mem::MaybeUninit;
+        use bpf::*;
+        use libbpf_rs::skel::SkelBuilder;
+        use libbpf_rs::skel::OpenSkel;
+
+        fn main() {
+            let builder = ProgSkelBuilder::default();
+            let mut open_object = MaybeUninit::uninit();
+            let _skel = builder
+                .open(&mut open_object)
+                .expect("failed to open skel")
+                .load()
+                .expect("failed to load skel");
+        }
+    "#
+    .to_string();
+    let () = build_rust_project_from_bpf_c(&bpf_c, &rust);
+}
+
+#[test]
+fn test_skeleton_basic() {
+    let bpf_c = r#"
+        #include "vmlinux.h"
+        #include <bpf/bpf_helpers.h>
+
+        struct {
+                __uint(type, BPF_MAP_TYPE_HASH);
+                __uint(max_entries, 1024);
+                __type(key, u32);
+                __type(value, u64);
+        } mymap SEC(".maps");
+
+        SEC("kprobe/foo")
+        int this_is_my_prog(u64 *ctx)
+        {
+                return 0;
+        }
+    "#
+    .to_string();
+    let rust = r#"
+        #![warn(elided_lifetimes_in_paths)]
+        mod bpf;
+        use std::mem::MaybeUninit;
         use bpf::*;
         use libbpf_rs::skel::SkelBuilder;
         use libbpf_rs::skel::OpenSkel;
         use libbpf_rs::skel::Skel;
 
-        fn main() {{
+        fn main() {
             let builder = ProgSkelBuilder::default();
-            let mut open_skel = builder
-                .open()
+            let mut open_object = MaybeUninit::uninit();
+            let open_skel = builder
+                .open(&mut open_object)
                 .expect("failed to open skel");
 
             // Check that we can grab handles to open maps/progs
-            let _open_map = open_skel.maps().mymap();
-            let _open_prog = open_skel.progs().this_is_my_prog();
-            let _open_map_mut = open_skel.maps_mut().mymap();
-            let _open_prog_mut = open_skel.progs_mut().this_is_my_prog();
+            let _open_map = &open_skel.maps.mymap;
+            let _open_prog = &open_skel.progs.this_is_my_prog;
 
-            let mut skel = open_skel
-                .load()
-                .expect("failed to load skel");
+            let mut skel = open_skel.load().expect("failed to load skel");
 
             // Check that we can grab handles to loaded maps/progs
-            let _map = skel.maps().mymap();
-            let _prog = skel.progs().this_is_my_prog();
-            let _map_mut = skel.maps_mut().mymap();
-            let _prog_mut = skel.progs_mut().this_is_my_prog();
+            let _map = &skel.maps.mymap;
+            let _prog = &skel.progs.this_is_my_prog;
 
             // Check that attach() is generated
             skel.attach().expect("failed to attach progs");
 
             // Check that Option<Link> field is generated
             let _mylink = skel.links.this_is_my_prog.unwrap();
-        }}
-        "#,
-    )
-    .expect("failed to write to main.rs");
+        }
+    "#
+    .to_string();
 
-    let status = Command::new("cargo")
-        .arg("build")
-        .arg("--quiet")
-        .arg("--manifest-path")
-        .arg(cargo_toml.into_os_string())
-        .env("RUSTFLAGS", "-Dwarnings")
-        .status()
-        .expect("failed to spawn cargo-build");
-    assert!(status.success());
+    let () = build_rust_project_from_bpf_c(&bpf_c, &rust);
 }
 
 #[test]
 fn test_skeleton_generate_datasec_static() {
-    let (_dir, proj_dir, cargo_toml) = setup_temp_project();
-
-    // Add prog dir
-    create_dir(proj_dir.join("src/bpf")).expect("failed to create prog dir");
-
-    // Add a prog
-    let mut prog = OpenOptions::new()
-        .write(true)
-        .create(true)
-        .truncate(true)
-        .open(proj_dir.join("src/bpf/prog.bpf.c"))
-        .expect("failed to open prog.bpf.c");
-
-    write!(
-        prog,
-        r#"
+    let bpf_c = r#"
         #include "vmlinux.h"
         #include <bpf/bpf_helpers.h>
 
         SEC("kprobe/foo")
         int this_is_my_prog(u64 *ctx)
-        {{
+        {
                 bpf_printk("this should not cause an error");
                 return 0;
-        }}
-        "#,
-    )
-    .expect("failed to write prog.bpf.c");
+        }
+    "#
+    .to_string();
 
-    // Lay down the necessary header files
-    add_vmlinux_header(&proj_dir);
+    let rust = r#"
+        #![warn(elided_lifetimes_in_paths)]
+        mod bpf;
+        use bpf::*;
 
-    make(
-        true,
-        Some(&cargo_toml),
-        None,
-        Vec::new(),
-        true,
-        true,
-        Vec::new(),
-        None,
-    )
-    .unwrap();
+        fn main() {
+            let _builder = ProgSkelBuilder::default();
+        }
+    "#
+    .to_string();
+    let () = build_rust_project_from_bpf_c(&bpf_c, &rust);
 }
 
 #[test]
 fn test_skeleton_datasec() {
-    let (_dir, proj_dir, cargo_toml) = setup_temp_project();
-
-    // Add prog dir
-    create_dir(proj_dir.join("src/bpf")).expect("failed to create prog dir");
-
-    // Add a prog
-    let mut prog = OpenOptions::new()
-        .write(true)
-        .create(true)
-        .truncate(true)
-        .open(proj_dir.join("src/bpf/prog.bpf.c"))
-        .expect("failed to open prog.bpf.c");
-
-    write!(
-        prog,
-        r#"
+    let bpf_c = r#"
         #include "vmlinux.h"
         #include <bpf/bpf_helpers.h>
 
@@ -654,100 +569,51 @@ fn test_skeleton_datasec() {
 
         SEC("kprobe/foo")
         int this_is_my_prog(u64 *ctx)
-        {{
+        {
                 return 0;
-        }}
-        "#,
-    )
-    .expect("failed to write prog.bpf.c");
+        }
+    "#
+    .to_string();
 
-    // Lay down the necessary header files
-    add_vmlinux_header(&proj_dir);
-
-    make(
-        true,
-        Some(&cargo_toml),
-        None,
-        Vec::new(),
-        true,
-        true,
-        Vec::new(),
-        None,
-    )
-    .unwrap();
-
-    let mut cargo = OpenOptions::new()
-        .append(true)
-        .open(&cargo_toml)
-        .expect("failed to open Cargo.toml");
-
-    // Make test project use our development libbpf-rs version
-    writeln!(
-        cargo,
-        r#"
-        libbpf-rs = {{ path = "{}" }}
-        "#,
-        get_libbpf_rs_path().as_path().display()
-    )
-    .expect("failed to write to Cargo.toml");
-
-    let mut source = OpenOptions::new()
-        .write(true)
-        .truncate(true)
-        .open(proj_dir.join("src/main.rs"))
-        .expect("failed to open main.rs");
-
-    write!(
-        source,
-        r#"
+    let rust = r#"
         #![warn(elided_lifetimes_in_paths)]
         mod bpf;
+        use std::mem::MaybeUninit;
         use bpf::*;
         use libbpf_rs::skel::SkelBuilder;
         use libbpf_rs::skel::OpenSkel;
 
-        fn main() {{
+        fn main() {
             let builder = ProgSkelBuilder::default();
-            let mut open_skel = builder
-                .open()
+            let mut open_object = MaybeUninit::uninit();
+            let open_skel = builder
+                .open(&mut open_object)
                 .expect("failed to open skel");
 
             // Check that we set rodata vars before load
-            open_skel.rodata_mut().myconst = std::ptr::null_mut();
+            open_skel.maps.rodata_data.myconst = std::ptr::null_mut();
 
             // We can always set bss vars
-            open_skel.bss_mut().myglobal = 42;
+            open_skel.maps.bss_data.myglobal = 42;
 
-            open_skel.data_custom_mut().mycustomdata = 1337;
-            open_skel.bss_custom_mut().mycustombss = 12;
-            assert_eq!(open_skel.rodata_custom_1().mycustomrodata, 43);
+            open_skel.maps.data_custom_data.mycustomdata = 1337;
+            open_skel.maps.bss_custom_data.mycustombss = 12;
+            assert_eq!(open_skel.maps.rodata_custom_1_data.mycustomrodata, 43);
 
-            let mut skel = open_skel
-                .load()
-                .expect("failed to load skel");
+            let skel = open_skel.load().expect("failed to load skel");
 
             // We can always set bss vars
-            skel.bss_mut().myglobal = 24;
-            skel.data_custom_mut().mycustomdata += 1;
-            skel.bss_custom_mut().mycustombss += 1;
-            assert_eq!(skel.rodata_custom_1().mycustomrodata, 43);
+            skel.maps.bss_data.myglobal = 24;
+            skel.maps.data_custom_data.mycustomdata += 1;
+            skel.maps.bss_custom_data.mycustombss += 1;
+            assert_eq!(skel.maps.rodata_custom_1_data.mycustomrodata, 43);
 
             // Read only for rodata after load
-            let _rodata: &prog_types::rodata = skel.rodata();
-        }}
-        "#,
-    )
-    .expect("failed to write to main.rs");
-
-    let status = Command::new("cargo")
-        .arg("build")
-        .arg("--quiet")
-        .arg("--manifest-path")
-        .arg(cargo_toml.into_os_string())
-        .env("RUSTFLAGS", "-Dwarnings")
-        .status()
-        .expect("failed to spawn cargo-build");
-    assert!(status.success());
+            let _rodata: &types::rodata = skel.maps.rodata_data;
+        }
+    "#
+    .to_string();
+    let () = build_rust_project_from_bpf_c(&bpf_c, &rust);
 }
 
 #[test]
@@ -771,13 +637,26 @@ fn test_skeleton_builder_basic() {
         #include "vmlinux.h"
         #include <bpf/bpf_helpers.h>
 
+        struct unique_key {{
+            int cap;
+            u32 tgid;
+            u64 cgroupid;
+        }};
+
         struct {{
                 __uint(type, BPF_MAP_TYPE_HASH);
                 __uint(max_entries, 1024);
-                __type(key, u32);
+                __type(key, struct unique_key);
                 __type(value, u64);
         }} mymap SEC(".maps");
 
+        struct {{
+                __uint(type, BPF_MAP_TYPE_HASH);
+                __uint(max_entries, 1024);
+                __uint(key_size, 8);
+                __type(value, u64);
+        }} mymap2 SEC(".maps");
+
         SEC("kprobe/foo")
         int this_is_my_prog(u64 *ctx)
         {{
@@ -824,6 +703,7 @@ fn test_skeleton_builder_basic() {
         r#"
         #[path = "{skel_path}"]
         mod skel;
+        use std::mem::MaybeUninit;
         use skel::*;
         use libbpf_rs::skel::SkelBuilder;
         use libbpf_rs::skel::OpenSkel;
@@ -831,32 +711,30 @@ fn test_skeleton_builder_basic() {
 
         fn main() {{
             let builder = ProgSkelBuilder::default();
-            let mut open_skel = builder
-                .open()
+            let mut open_object = MaybeUninit::uninit();
+            let open_skel = builder
+                .open(&mut open_object)
                 .expect("failed to open skel");
 
             // Check that we can grab handles to open maps/progs
-            let _open_map = open_skel.maps().mymap();
-            let _open_prog = open_skel.progs().this_is_my_prog();
-            let _open_map_mut = open_skel.maps_mut().mymap();
-            let _open_prog_mut = open_skel.progs_mut().this_is_my_prog();
+            let _open_map = &open_skel.maps.mymap;
+            let _open_map2 = &open_skel.maps.mymap2;
+            let _open_prog = &open_skel.progs.this_is_my_prog;
 
-
-            let mut skel = open_skel
-                .load()
-                .expect("failed to load skel");
+            let mut skel = open_skel.load().expect("failed to load skel");
 
             // Check that we can grab handles to loaded maps/progs
-            let _map = skel.maps().mymap();
-            let _prog = skel.progs().this_is_my_prog();
-            let _map_mut = skel.maps_mut().mymap();
-            let _prog_mut = skel.progs_mut().this_is_my_prog();
+            let _map = &skel.maps.mymap;
+            let _map2 = &skel.maps.mymap2;
+            let _prog = &skel.progs.this_is_my_prog;
 
             // Check that attach() is generated
             skel.attach().expect("failed to attach progs");
 
             // Check that Option<Link> field is generated
             let _mylink = skel.links.this_is_my_prog.unwrap();
+
+            let _key = types::unique_key::default();
         }}
         "#,
         skel_path = skel.path().display(),
@@ -921,209 +799,232 @@ fn test_skeleton_builder_clang_opts() {
 
 #[test]
 fn test_skeleton_builder_arrays_ptrs() {
-    let (_dir, proj_dir, cargo_toml) = setup_temp_project();
-
-    // Add prog dir
-    create_dir(proj_dir.join("src/bpf")).expect("failed to create prog dir");
-
-    // Add a prog
-    let mut prog = OpenOptions::new()
-        .write(true)
-        .create(true)
-        .truncate(true)
-        .open(proj_dir.join("src/bpf/prog.bpf.c"))
-        .expect("failed to open prog.bpf.c");
-
-    write!(
-        prog,
-        r#"
+    let bpf_c = r#"
         #include "vmlinux.h"
         #include <bpf/bpf_helpers.h>
 
-        struct inner {{
+        struct inner {
             int a;
-        }};
+        };
 
-        struct mystruct {{
+        struct mystruct {
             int x;
-            struct {{
+            struct {
                 int b;
-            }} y[2];
+            } y[2];
             struct inner z[2];
-        }};
+        };
 
-        const volatile struct mystruct my_array[1] = {{ {{0}} }};
+        const volatile struct mystruct my_array[1] = { {0} };
         struct mystruct * const my_ptr = NULL;
-        "#,
-    )
-    .expect("failed to write prog.bpf.c");
+    "#
+    .to_string();
 
-    // Lay down the necessary header files
-    add_vmlinux_header(&proj_dir);
-
-    make(
-        true,
-        Some(&cargo_toml),
-        None,
-        Vec::new(),
-        true,
-        true,
-        Vec::new(),
-        None,
-    )
-    .unwrap();
-
-    let mut cargo = OpenOptions::new()
-        .append(true)
-        .open(&cargo_toml)
-        .expect("failed to open Cargo.toml");
-
-    // Make test project use our development libbpf-rs version
-    writeln!(
-        cargo,
-        r#"
-        libbpf-rs = {{ path = "{}" }}
-        "#,
-        get_libbpf_rs_path().as_path().display()
-    )
-    .expect("failed to write to Cargo.toml");
-
-    let mut source = OpenOptions::new()
-        .write(true)
-        .truncate(true)
-        .open(proj_dir.join("src/main.rs"))
-        .expect("failed to open main.rs");
-
-    write!(
-        source,
-        r#"
+    let rust = r#"
         #![warn(elided_lifetimes_in_paths)]
         mod bpf;
+        use std::mem::MaybeUninit;
         use bpf::*;
         use libbpf_rs::skel::SkelBuilder;
 
-        fn main() {{
+        fn main() {
             let builder = ProgSkelBuilder::default();
+            let mut open_object = MaybeUninit::uninit();
             let open_skel = builder
-                .open()
+                .open(&mut open_object)
                 .expect("failed to open skel");
 
             // That everything exists and compiled okay
-            let _ = open_skel.rodata().my_array[0].x;
-            let _ = open_skel.rodata().my_array[0].y[1].b;
-            let _ = open_skel.rodata().my_array[0].z[0].a;
-            let _ = open_skel.rodata().my_ptr;
-        }}
-        "#,
-    )
-    .expect("failed to write to main.rs");
-
-    let status = Command::new("cargo")
-        .arg("build")
-        .arg("--quiet")
-        .arg("--manifest-path")
-        .arg(cargo_toml.into_os_string())
-        .env("RUSTFLAGS", "-Dwarnings")
-        .status()
-        .expect("failed to spawn cargo-build");
-
-    assert!(status.success());
+            let _ = open_skel.maps.rodata_data.my_array[0].x;
+            let _ = open_skel.maps.rodata_data.my_array[0].y[1].b;
+            let _ = open_skel.maps.rodata_data.my_array[0].z[0].a;
+            let _ = open_skel.maps.rodata_data.my_ptr;
+        }
+    "#
+    .to_string();
+    let () = build_rust_project_from_bpf_c(&bpf_c, &rust);
 }
 
 #[test]
 fn test_skeleton_generate_struct_with_pointer() {
-    let (_dir, proj_dir, cargo_toml) = setup_temp_project();
-
-    // Add prog dir
-    create_dir(proj_dir.join("src/bpf")).expect("failed to create prog dir");
-
-    // Add a prog
-    let mut prog = OpenOptions::new()
-        .write(true)
-        .create(true)
-        .truncate(true)
-        .open(proj_dir.join("src/bpf/prog.bpf.c"))
-        .expect("failed to open prog.bpf.c");
-
-    write!(
-        prog,
-        r#"
+    let bpf_c = r#"
         #include "vmlinux.h"
         #include <bpf/bpf_helpers.h>
 
-        struct list {{
+        struct list {
             struct list *next;
-        }};
+        };
 
         struct list l;
-        "#,
-    )
-    .expect("failed to write prog.bpf.c");
+    "#
+    .to_string();
 
-    // Lay down the necessary header files
-    add_vmlinux_header(&proj_dir);
-
-    make(
-        true,
-        Some(&cargo_toml),
-        None,
-        Vec::new(),
-        true,
-        true,
-        Vec::new(),
-        None,
-    )
-    .unwrap();
-
-    let mut cargo = OpenOptions::new()
-        .append(true)
-        .open(&cargo_toml)
-        .expect("failed to open Cargo.toml");
-
-    // Make test project use our development libbpf-rs version
-    writeln!(
-        cargo,
-        r#"
-        libbpf-rs = {{ path = "{}" }}
-        "#,
-        get_libbpf_rs_path().as_path().display()
-    )
-    .expect("failed to write to Cargo.toml");
-
-    let mut source = OpenOptions::new()
-        .write(true)
-        .truncate(true)
-        .open(proj_dir.join("src/main.rs"))
-        .expect("failed to open main.rs");
-
-    write!(
-        source,
-        r#"
+    let rust = r#"
         #![warn(elided_lifetimes_in_paths)]
         mod bpf;
+        use std::mem::MaybeUninit;
         use bpf::*;
         use libbpf_rs::skel::SkelBuilder;
 
-        fn main() {{
+        fn main() {
             let builder = ProgSkelBuilder::default();
+            let mut open_object = MaybeUninit::uninit();
             let _open_skel = builder
-                .open()
+                .open(&mut open_object)
                 .expect("failed to open skel");
-        }}
-        "#,
-    )
-    .expect("failed to write to main.rs");
+        }
+    "#
+    .to_string();
+    let () = build_rust_project_from_bpf_c(&bpf_c, &rust);
+}
 
-    let status = Command::new("cargo")
-        .arg("build")
-        .arg("--quiet")
-        .arg("--manifest-path")
-        .arg(cargo_toml.into_os_string())
-        .env("RUSTFLAGS", "-Dwarnings")
-        .status()
-        .expect("failed to spawn cargo-build");
+/// Check that we generate valid Rust code for an array of pointers.
+#[test]
+fn test_skeleton_generate_struct_with_pointer_array() {
+    let bpf_c = r#"
+        #include "vmlinux.h"
+        #include <bpf/bpf_helpers.h>
 
-    assert!(status.success());
+        struct vm_area_struct;
+
+        struct vmacache {
+            struct vm_area_struct *vmas[4];
+        };
+
+        struct vmacache c;
+    "#
+    .to_string();
+
+    let rust = r#"
+        #![warn(elided_lifetimes_in_paths)]
+        mod bpf;
+        use std::mem::MaybeUninit;
+        use bpf::*;
+        use libbpf_rs::skel::SkelBuilder;
+
+        fn main() {
+            let builder = ProgSkelBuilder::default();
+            let mut open_object = MaybeUninit::uninit();
+            let _open_skel = builder
+                .open(&mut open_object)
+                .expect("failed to open skel");
+        }
+    "#
+    .to_string();
+    let () = build_rust_project_from_bpf_c(&bpf_c, &rust);
+}
+
+/// Generate a skeleton that includes multiple "anon" type definitions.
+#[test]
+fn test_skeleton_builder_multiple_anon() {
+    let bpf_c = r#"
+        #include "vmlinux.h"
+        #include <bpf/bpf_helpers.h>
+
+        struct unique_key {
+            struct {
+                u32 foo;
+                u64 bar;
+            } foobar;
+        };
+
+        struct {
+                __uint(type, BPF_MAP_TYPE_HASH);
+                __uint(max_entries, 1024);
+                __type(key, struct unique_key);
+                __type(value, u64);
+        } mymap SEC(".maps");
+
+        struct Foo {
+            int x;
+            struct {
+                u8 y[10];
+                u16 z[16];
+            } bar;
+            struct {
+                u32 w;
+                u64 *u;
+            } baz;
+            int w;
+        };
+
+        struct Foo foo;
+
+        SEC("kprobe/foo")
+        int this_is_my_prog(u64 *ctx)
+        {
+                return 0;
+        }
+    "#
+    .to_string();
+
+    let rust = r#"
+        mod bpf;
+        use std::mem::MaybeUninit;
+        use bpf::*;
+        use libbpf_rs::skel::SkelBuilder;
+        use libbpf_rs::skel::OpenSkel;
+
+        fn main() {
+            let builder = ProgSkelBuilder::default();
+            let mut open_object = MaybeUninit::uninit();
+            let open_skel = builder
+                .open(&mut open_object)
+                .expect("failed to open skel");
+
+            let _skel = open_skel.load().expect("failed to load skel");
+            let _key = types::unique_key::default();
+        }
+    "#
+    .to_string();
+
+    let () = build_rust_project_from_bpf_c(&bpf_c, &rust);
+}
+
+/// Test that we generate a valid skeleton when multiple kfuncs are being used.
+#[test]
+fn test_skeleton_multipl_kfuncs() {
+    let bpf_c = r#"
+        #include "vmlinux.h"
+        #include <bpf/bpf_helpers.h>
+
+        extern int bpf_dynptr_from_xdp(
+            struct xdp_md* xdp,
+            __u64 flags,
+            struct bpf_dynptr* ptr__uninit) __ksym;
+
+        extern void* bpf_dynptr_slice(
+            const struct bpf_dynptr* ptr,
+            __u32 offset,
+            void* buffer,
+            __u32 buffer__szk) __ksym;
+
+        SEC("xdp")
+        int steering(struct xdp_md* xdp) {
+            u8 ethb[sizeof(struct ethhdr)];
+            struct bpf_dynptr ptr;
+            struct ethhdr* eth;
+
+            if (bpf_dynptr_from_xdp(xdp, 0, &ptr)) {
+              return XDP_PASS;
+            }
+            eth = bpf_dynptr_slice(&ptr, 0, ethb, sizeof(ethb));
+            (void)eth;
+        }
+    "#
+    .to_string();
+
+    let rust = r#"
+        #![warn(elided_lifetimes_in_paths)]
+        mod bpf;
+        use bpf::*;
+
+        fn main() {
+            let _builder = ProgSkelBuilder::default();
+        }
+    "#
+    .to_string();
+    let () = build_rust_project_from_bpf_c(&bpf_c, &rust);
 }
 
 /// Check that skeleton creation is deterministic, i.e., that no temporary paths
@@ -1331,6 +1232,7 @@ fn test_btf_dump_reserved_keyword_escaping() {
 struct Foo {
     u64 type;
     void* mod;
+    void* self;
 };
 
 struct Foo foo = {{0}};
@@ -1342,12 +1244,14 @@ struct Foo foo = {{0}};
 pub struct Foo {
     pub r#type: u64,
     pub r#mod: *mut std::ffi::c_void,
+    pub slf: *mut std::ffi::c_void,
 }
 impl Default for Foo {
     fn default() -> Self {
-        Foo {
+        Self {
             r#type: u64::default(),
             r#mod: std::ptr::null_mut(),
+            slf: std::ptr::null_mut(),
         }
     }
 }
@@ -1387,7 +1291,7 @@ pub struct Foo {
 }
 impl Default for Foo {
     fn default() -> Self {
-        Foo {
+        Self {
             x: i32::default(),
             y: [i8::default(); 10],
             z: std::ptr::null_mut(),
@@ -1467,6 +1371,43 @@ pub struct Foo {
     assert_definition(&btf, &struct_foo, expected_output);
 }
 
+#[test]
+fn test_btf_dump_insane_align() {
+    let prog_text = r#"
+#include "vmlinux.h"
+#include <bpf/bpf_helpers.h>
+
+struct Foo {
+    int x;
+} __attribute__((aligned(64)));
+
+struct Foo foo = {1};
+"#;
+
+    let expected_output = r#"
+#[derive(Debug, Copy, Clone)]
+#[repr(C)]
+pub struct Foo {
+    pub x: i32,
+    pub __pad_4: [u8; 60],
+}
+impl Default for Foo {
+    fn default() -> Self {
+        Self {
+            x: i32::default(),
+            __pad_4: [u8::default(); 60],
+        }
+    }
+}
+"#;
+
+    let mmap = build_btf_mmap(prog_text);
+    let btf = btf_from_mmap(&mmap);
+    let struct_foo = find_type_in_btf!(btf, types::Struct<'_>, "Foo");
+
+    assert_definition(&btf, &struct_foo, expected_output);
+}
+
 #[test]
 fn test_btf_dump_basic_long_array() {
     let prog_text = r#"
@@ -1494,7 +1435,7 @@ pub struct Foo {
 }
 impl Default for Foo {
     fn default() -> Self {
-        Foo {
+        Self {
             x: i32::default(),
             y: [i8::default(); 33],
             z: std::ptr::null_mut(),
@@ -1565,7 +1506,7 @@ pub struct Foo {
 }
 impl Default for Foo {
     fn default() -> Self {
-        Foo {
+        Self {
             ip: std::ptr::null_mut(),
             ipp: std::ptr::null_mut(),
             bar: Bar::default(),
@@ -1600,8 +1541,8 @@ fn test_btf_dump_struct_definition_func_proto() {
 #include <bpf/bpf_helpers.h>
 
 struct with_func_proto {
-	struct with_func_proto *next;
-	void (*func)(struct with_func_proto *);
+    struct with_func_proto *next;
+    void (*func)(struct with_func_proto *);
 };
 
 struct with_func_proto w;
@@ -1616,7 +1557,7 @@ pub struct with_func_proto {
 }
 impl Default for with_func_proto {
     fn default() -> Self {
-        with_func_proto {
+        Self {
             next: std::ptr::null_mut(),
             func: std::ptr::null_mut(),
         }
@@ -1674,7 +1615,7 @@ pub struct Foo {
 }
 impl Default for Foo {
     fn default() -> Self {
-        Foo {
+        Self {
             ip: std::ptr::null_mut(),
             ipp: std::ptr::null_mut(),
             bar: Bar::default(),
@@ -1694,7 +1635,7 @@ pub struct Bar {
 }
 impl Default for Bar {
     fn default() -> Self {
-        Bar {
+        Self {
             x: u16::default(),
             y: [u16::default(); 33],
         }
@@ -1770,7 +1711,7 @@ pub struct Foo {
 }
 impl Default for Foo {
     fn default() -> Self {
-        Foo {
+        Self {
             x: i32::default(),
             y: i8::default(),
             z: [i32::default(); 33],
@@ -1849,7 +1790,7 @@ pub struct Foo {
 }
 impl Default for Foo {
     fn default() -> Self {
-        Foo {
+        Self {
             name: std::ptr::null_mut(),
             kset: std::ptr::null_mut(),
             __pad_16: [u8::default(); 1],
@@ -1932,7 +1873,7 @@ impl std::fmt::Debug for Foo {
 }
 impl Default for Foo {
     fn default() -> Self {
-        Foo {
+        Self {
             x: i32::default(),
         }
     }
@@ -2021,7 +1962,7 @@ pub struct Foo {
 }
 impl Default for Foo {
     fn default() -> Self {
-        Foo {
+        Self {
             x: i32::default(),
             y: [i8::default(); 10],
             z: std::ptr::null_mut(),
@@ -2127,7 +2068,7 @@ pub struct Foo {
 }
 impl Default for Foo {
     fn default() -> Self {
-        Foo {
+        Self {
             x: i32::default(),
             y: [i8::default(); 33],
             z: std::ptr::null_mut(),
@@ -2193,7 +2134,7 @@ pub struct Foo {
 }
 impl Default for Foo {
     fn default() -> Self {
-        Foo {
+        Self {
             x: i32::default(),
             y: [i8::default(); 10],
             z: std::ptr::null_mut(),
@@ -2261,7 +2202,7 @@ pub struct Foo {
 }
 impl Default for Foo {
     fn default() -> Self {
-        Foo {
+        Self {
             x: i32::default(),
             y: [i8::default(); 33],
             z: std::ptr::null_mut(),
@@ -2337,7 +2278,7 @@ impl std::fmt::Debug for __anon_1 {
 }
 impl Default for __anon_1 {
     fn default() -> Self {
-        __anon_1 {
+        Self {
             y: [u8::default(); 10],
         }
     }
@@ -2355,7 +2296,7 @@ impl std::fmt::Debug for __anon_2 {
 }
 impl Default for __anon_2 {
     fn default() -> Self {
-        __anon_2 {
+        Self {
             w: u32::default(),
         }
     }
@@ -2418,7 +2359,7 @@ pub struct __anon_2 {
 }
 impl Default for __anon_2 {
     fn default() -> Self {
-        __anon_2 {
+        Self {
             w: u32::default(),
             __pad_4: [u8::default(); 4],
             u: std::ptr::null_mut(),
@@ -2497,7 +2438,7 @@ impl std::fmt::Debug for __anon_2 {
 }
 impl Default for __anon_2 {
     fn default() -> Self {
-        __anon_2 {
+        Self {
             a: std::ptr::null_mut(),
         }
     }
@@ -2511,7 +2452,7 @@ pub struct __anon_3 {
 }
 impl Default for __anon_3 {
     fn default() -> Self {
-        __anon_3 {
+        Self {
             w: u32::default(),
             __pad_4: [u8::default(); 4],
             u: std::ptr::null_mut(),
@@ -2531,7 +2472,7 @@ impl std::fmt::Debug for __anon_4 {
 }
 impl Default for __anon_4 {
     fn default() -> Self {
-        __anon_4 {
+        Self {
             c: u8::default(),
         }
     }
@@ -2582,7 +2523,7 @@ impl std::fmt::Debug for __anon_1 {
 }
 impl Default for __anon_1 {
     fn default() -> Self {
-        __anon_1 {
+        Self {
             name: std::ptr::null_mut(),
         }
     }
@@ -2619,7 +2560,7 @@ pub struct Foo {
 }
 impl Default for Foo {
     fn default() -> Self {
-        Foo {
+        Self {
             test: std::mem::MaybeUninit::new(__anon_1::default()),
         }
     }
@@ -2669,7 +2610,7 @@ pub struct Foo {
 }
 impl Default for Foo {
     fn default() -> Self {
-        Foo {
+        Self {
             a: i32::default(),
             b: u16::default(),
             c: i16::default(),
@@ -2698,20 +2639,20 @@ fn test_btf_dump_definition_unnamed_union() {
 // re-typed 'struct bpf_sock_tuple tup' from vmlinux as of kernel 5.15
 // with a little bit added for additional complexity testing
 struct bpf_sock_tuple_5_15 {
-	union {
-		struct {
-			__be32 saddr;
-			__be32 daddr;
-			__be16 sport;
-			__be16 dport;
-		} ipv4;
-		struct {
-			__be32 saddr[4];
-			__be32 daddr[4];
-			__be16 sport;
-			__be16 dport;
-		} ipv6;
-	};
+    union {
+        struct {
+            __be32 saddr;
+            __be32 daddr;
+            __be16 sport;
+            __be16 dport;
+        } ipv4;
+        struct {
+            __be32 saddr[4];
+            __be32 daddr[4];
+            __be16 sport;
+            __be16 dport;
+        } ipv6;
+    };
 
     union {
         int a;
@@ -2742,7 +2683,7 @@ impl std::fmt::Debug for __anon_1 {
 }
 impl Default for __anon_1 {
     fn default() -> Self {
-        __anon_1 {
+        Self {
             ipv4: __anon_3::default(),
         }
     }
@@ -2760,7 +2701,7 @@ impl std::fmt::Debug for __anon_2 {
 }
 impl Default for __anon_2 {
     fn default() -> Self {
-        __anon_2 {
+        Self {
             a: i32::default(),
         }
     }
@@ -2792,6 +2733,36 @@ pub struct __anon_4 {
     assert_definition(&btf, &struct_bpf_sock_tuple, expected_output);
 }
 
+#[test]
+fn test_btf_dump_definition_empty_union() {
+    let prog_text = r#"
+#include "vmlinux.h"
+
+struct struct_with_empty_union {
+    int member;
+    union {};
+};
+struct struct_with_empty_union s;
+"#;
+
+    let expected_output = r#"
+#[derive(Debug, Default, Copy, Clone)]
+#[repr(C)]
+pub struct struct_with_empty_union {
+    pub member: i32,
+}
+"#;
+
+    let mmap = build_btf_mmap(prog_text);
+    let btf = btf_from_mmap(&mmap);
+
+    // Find the struct
+    let struct_with_empty_union =
+        find_type_in_btf!(btf, types::Struct<'_>, "struct_with_empty_union");
+
+    assert_definition(&btf, &struct_with_empty_union, expected_output);
+}
+
 #[test]
 fn test_btf_dump_definition_struct_ops_mixed() {
     let prog_text = r#"
@@ -2802,49 +2773,49 @@ fn test_btf_dump_definition_struct_ops_mixed() {
 SEC("struct_ops/test_1")
 int BPF_PROG(test_1, struct bpf_dummy_ops_state *state)
 {
-	return 0;
+    return 0;
 }
 
 SEC(".struct_ops")
 struct bpf_dummy_ops dummy_1 = {
-	.test_1 = (void *)test_1,
+    .test_1 = (void *)test_1,
 };
 
 SEC(".struct_ops.link")
 struct bpf_dummy_ops dummy_2 = {
-	.test_1 = (void *)test_1,
+    .test_1 = (void *)test_1,
 };
 "#;
 
     let expected_output = r#"
 #[derive(Debug, Clone)]
 #[repr(C)]
-pub struct struct_ops {
-    pub dummy_1: *mut bpf_dummy_ops,
-    pub dummy_2: *mut bpf_dummy_ops,
+pub struct StructOps {
+    pub dummy_1: *mut types::bpf_dummy_ops,
+    pub dummy_2: *mut types::bpf_dummy_ops,
 }
 
-impl struct_ops {
+impl StructOps {
 
-    pub fn dummy_1(&self) -> &bpf_dummy_ops {
+    pub fn dummy_1(&self) -> &types::bpf_dummy_ops {
         // SAFETY: The library ensures that the member is pointing to
         //         valid data.
         unsafe { self.dummy_1.as_ref() }.unwrap()
     }
 
-    pub fn dummy_1_mut(&mut self) -> &mut bpf_dummy_ops {
+    pub fn dummy_1_mut(&mut self) -> &mut types::bpf_dummy_ops {
         // SAFETY: The library ensures that the member is pointing to
         //         valid data.
         unsafe { self.dummy_1.as_mut() }.unwrap()
     }
 
-    pub fn dummy_2(&self) -> &bpf_dummy_ops {
+    pub fn dummy_2(&self) -> &types::bpf_dummy_ops {
         // SAFETY: The library ensures that the member is pointing to
         //         valid data.
         unsafe { self.dummy_2.as_ref() }.unwrap()
     }
 
-    pub fn dummy_2_mut(&mut self) -> &mut bpf_dummy_ops {
+    pub fn dummy_2_mut(&mut self) -> &mut types::bpf_dummy_ops {
         // SAFETY: The library ensures that the member is pointing to
         //         valid data.
         unsafe { self.dummy_2.as_mut() }.unwrap()
@@ -2855,12 +2826,14 @@ impl struct_ops {
 pub struct bpf_dummy_ops {
     pub test_1: *mut libbpf_rs::libbpf_sys::bpf_program,
     pub test_2: *mut libbpf_rs::libbpf_sys::bpf_program,
+    pub test_sleepable: *mut libbpf_rs::libbpf_sys::bpf_program,
 }
 impl Default for bpf_dummy_ops {
     fn default() -> Self {
-        bpf_dummy_ops {
+        Self {
             test_1: std::ptr::null_mut(),
             test_2: std::ptr::null_mut(),
+            test_sleepable: std::ptr::null_mut(),
         }
     }
 }
@@ -2869,7 +2842,10 @@ impl Default for bpf_dummy_ops {
     let mmap = build_btf_mmap(prog_text);
     let btf = btf_from_mmap(&mmap);
     let mut processed = HashSet::new();
-    let def = btf.struct_ops_type_definition(&mut processed).unwrap();
+    let mut def = String::new();
+    let gen = GenStructOps::new(&btf).unwrap();
+    let () = gen.gen_struct_ops_def(&mut def).unwrap();
+    let () = gen.gen_dependent_types(&mut processed, &mut def).unwrap();
 
     assert_output(&def, expected_output);
 }
